/*-
 * Copyright (c) 2000 Doug White
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* kqueuemodule.c: Python module for kqueue support.
 * Version 1.1 (Added copyright)
 *
 * Doug White
 * 
 */

#include <Python.h>
#include <sys/event.h>
#include <stdio.h>
#include <errno.h>

#define MAX_KEVENTS 512

/* Extract the element "element" from kelist from index i and return it. */

long
kqgetlistvalue(PyObject *kelist, int i, char* element)
{
  PyObject* temp;
  long t;
  temp = PyObject_GetAttrString(PyList_GetItem(kelist, i), element);
  if(temp == NULL) {
    /* Pass up exception */
    return NULL;
  }
  t = PyInt_AsLong(temp);
  Py_DECREF(temp); 
  return t;
}

/* Call kqueue(2) to create a new queue. */

static PyObject *
kqueue_kqueue(self, args)
     PyObject *self;
     PyObject *args;
{
  
  /* XXX BAD ASSUMPTION: kqueue() always works */
  return PyInt_FromLong((long)kqueue());
}

/* Call kevent(2) and do appropriate digestion of lists. */

static PyObject *
kqueue_kevent(self, args) 
     PyObject *self;
     PyObject *args;
{
  
  int kfd=0; 
  int wantNumEvents=0;
  int timeout=0;
  int haveNumEvents=0;
  int gotNumEvents=0;
  int i=0;
  PyObject *kelist, *temp, *output;
  struct kevent *event, *events;
  struct kevent *triggered;
  struct timespec totimespec;
  

  if(!PyArg_ParseTuple(args, "iOii", &kfd, &kelist, &wantNumEvents, &timeout)) 
  {
    return NULL;
  }
  
  if(!PyList_Check(kelist)) {
    PyErr_SetString(PyExc_TypeError, "Event list is not a list");
    return NULL; /* make sure this is a list */
  }

  /* Chop up kelist into appropriate C array. */
  /* kelist is a list of KEvent Python objects, so we'll have to dig
   * into it to get our parameters to load the kevent structs with. 
   */

  haveNumEvents = PyList_Size(kelist);

  /* If there's no events to process, don't bother. */
  if(haveNumEvents > 0) {

    if(!(events = malloc(sizeof(struct kevent) * haveNumEvents))) {
      PyErr_SetFromErrno(PyExc_MemoryError);
      return NULL;
    }
    
    for(i=0; i < haveNumEvents; i++) {

      event = &events[i];
      
      /* Somewhat dangerous assumptions here that these are defined. 
       * The Py_DECREF is necessary since PyObject_GetAttrString creates 
       * a reference to the item it gets from PyList_GetItem (which gets
       * a borrowed reference from the original list item).
       */

      /* XXX error checking?? */

      event->ident = (u_long)kqgetlistvalue(kelist, i, "ident");
      event->filter = (short)kqgetlistvalue(kelist, i, "filter");
      event->flags = (u_short)kqgetlistvalue(kelist, i, "flags");
      event->fflags = (u_int)kqgetlistvalue(kelist, i, "fflags");
      
      /* Opaque user data (probably a Python object */
      /* This is special so we can't use our helper function */
      event->udata = (void *)PyObject_GetAttrString(PyList_GetItem(kelist, i), "udata");
    }

  } else {
    events = NULL;
  }

  /* Allocate some space to hold the triggered events */
  if(!(triggered = malloc(sizeof(struct kevent*) * wantNumEvents))) {
    PyErr_SetFromErrno(PyExc_MemoryError);
    return NULL;
  }

  /* Build timespec for timeout */
  totimespec.tv_sec = timeout / 1000;
  totimespec.tv_nsec = (timeout % 1000) * 100000;

   printf("timespec: sec=%d nsec=%d\n", totimespec.tv_sec, totimespec.tv_nsec); 

  /* Make the call */
  gotNumEvents = kevent(kfd, events, haveNumEvents, triggered, wantNumEvents, &totimespec);

  /* Don't need the input event list anymore, so get rid of it */
  if(events != NULL) {
    free(events);
  }

  switch(gotNumEvents) {
    /* error */  
  case -1: 
    PyErr_SetFromErrno(PyExc_OSError);
    free(triggered);
    return NULL;
    break;

    /* timeout */
  case 0:
    /* return empty list */
    free(triggered);
    return Py_BuildValue("[]");
    break;

    /* Succeeded, got something back; return it in a list */
  default:
    if((output = PyList_New(0)) == NULL) return NULL;
    for(i=0; i < gotNumEvents; i++) 
      /* Do reverse of above, pick out of struct and load a tuple 
       * with the results.
       * Watch the reference counts on 'udata' .. The 'O' type will
       * increment it!
       */
      PyList_Append(output, Py_BuildValue("(i, i, i, i, i, N)", 
					  triggered[i].ident, 
					  triggered[i].filter,
					  triggered[i].flags,
					  triggered[i].fflags,
					  triggered[i].data,
					  triggered[i].udata));
  }
  free(triggered);
  /* pass back the results */
  return output;
  
}

/* Method table */

static PyMethodDef KQSyscallMethods[] = {
  { "kqueue", kqueue_kqueue, METH_VARARGS },
  { "kevent", kqueue_kevent, METH_VARARGS },
  { NULL, NULL }
};

void
initkqsyscall()
{
  (void) Py_InitModule("kqsyscall", KQSyscallMethods);
}
